package ru.dbotthepony.kstarbound.server.world

import it.unimi.dsi.fastutil.objects.Object2DoubleOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectAVLTreeSet
import kotlinx.coroutines.async
import kotlinx.coroutines.coroutineScope
import kotlinx.coroutines.future.await
import kotlinx.coroutines.launch
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.Globals
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.monster.MonsterVariant
import ru.dbotthepony.kstarbound.defs.tile.NO_DUNGEON_ID
import ru.dbotthepony.kstarbound.defs.tile.isEmptyTile
import ru.dbotthepony.kstarbound.defs.tile.isNotEmptyLiquid
import ru.dbotthepony.kstarbound.defs.world.SpawnArea
import ru.dbotthepony.kstarbound.defs.world.SpawnParameters
import ru.dbotthepony.kstarbound.defs.world.SpawnRegion
import ru.dbotthepony.kstarbound.defs.world.SpawnTime
import ru.dbotthepony.kstarbound.json.mergeJson
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.AABBi
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.math.vector.Vector2i
import ru.dbotthepony.kstarbound.util.random.nextRange
import ru.dbotthepony.kstarbound.util.random.shuffle
import ru.dbotthepony.kstarbound.util.random.staticRandom64
import ru.dbotthepony.kstarbound.util.random.staticRandomLong
import ru.dbotthepony.kstarbound.util.supplyAsync
import ru.dbotthepony.kstarbound.world.ChunkState
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import ru.dbotthepony.kstarbound.world.entities.MonsterEntity
import ru.dbotthepony.kstarbound.world.physics.CollisionType
import java.util.Comparator
import java.util.EnumSet
import java.util.concurrent.Future
import java.util.concurrent.TimeUnit
import kotlin.jvm.optionals.getOrNull
import kotlin.math.min

class MonsterSpawner(val world: ServerWorld) {
	var active = Globals.spawner.defaultActive

	private val activeCells = Object2DoubleOpenHashMap<Vector2i>()
	private val spawnedEntities = ArrayList<MonsterEntity>()
	private var hasCleanupTask = false

	private fun cell2region(cell: Vector2i): AABBi = AABBi(cell * Globals.spawner.spawnCellSize, (cell + Vector2i.POSITIVE_XY) * Globals.spawner.spawnCellSize)

	private fun getSpawnParameters(cell: Vector2i, chunkMap: World<*, *>.ChunkMap): SpawnParameters {
		var emptyCount = 0
		var nearSurfaceCount = 0
		var nearCeilingCount = 0
		var airCount = 0
		var liquidCount = 0
		var exposedCount = 0

		for (x in cell.x * Globals.spawner.spawnCellSize until (cell.x + 1) * Globals.spawner.spawnCellSize) {
			for (y in cell.y * Globals.spawner.spawnCellSize until (cell.y + 1) * Globals.spawner.spawnCellSize) {
				// Only empty blocks count towards spawn totals
				val cell = chunkMap.getCell(x, y)

				if (cell.foreground.material.value.collisionKind == CollisionType.NONE) {
					emptyCount++

					if (cell.liquid.state.isNotEmptyLiquid && cell.liquid.level > Globals.spawner.minimumLiquidLevel)
						liquidCount++

					if (cell.background.material.isEmptyTile)
						exposedCount++

					// The empty block will either count as an air block, a
					// "near-surface" block, or a "near-ceiling" block.  It will count as a
					// near-surface block if it is within the NearSurfaceDistance of a
					// CollsionKind::Block or CollisionKind::Platform block. If it is not a
					// near-surface block, it will count as a near-ceiling block if it is
					// within the NearCeilingDistance of a CollisionKind::Block.
					var nearSurface = false

					for (delta in 1 .. Globals.spawner.spawnCellNearSurfaceDistance) {
						if (chunkMap.getCell(x, y - delta).foreground.material.value.collisionKind.isFloorCollision) {
							nearSurface = true
							break
						}
					}

					var nearCeiling = false

					for (delta in 1 .. Globals.spawner.spawnCellNearCeilingDistance) {
						if (chunkMap.getCell(x, y + delta).foreground.material.value.collisionKind.isSolidCollision) {
							nearCeiling = true
							break
						}
					}

					if (nearSurface)
						nearSurfaceCount++
					else if (nearCeiling)
						nearCeilingCount++
					else
						airCount++
				}
			}
		}

		val spawnAreas = EnumSet.noneOf(SpawnArea::class.java)

		if (liquidCount > Globals.spawner.spawnCellMinimumLiquidTiles)
			spawnAreas.add(SpawnArea.LIQUID)
		if (nearSurfaceCount > Globals.spawner.spawnCellMinimumNearSurfaceTiles)
			spawnAreas.add(SpawnArea.SURFACE)
		if (nearCeilingCount > Globals.spawner.spawnCellMinimumNearCeilingTiles)
			spawnAreas.add(SpawnArea.CEILING)
		if (airCount > Globals.spawner.spawnCellMinimumAirTiles)
			spawnAreas.add(SpawnArea.AIR)
		if (emptyCount < Globals.spawner.spawnCellMinimumEmptyTiles)
			spawnAreas.add(SpawnArea.SOLID)

		val spawnRegion = if (exposedCount >= Globals.spawner.spawnCellMinimumExposedTiles)
			SpawnRegion.EXPOSED
		else
			SpawnRegion.ENCLOSED

		val time = if (world.sky.dayLevel >= Globals.spawner.minimumDayLevel)
			SpawnTime.DAY
		else
			SpawnTime.NIGHT

		return SpawnParameters(spawnAreas, spawnRegion, time)
	}

	private fun selectSpawnPosition(positions: Array<Vector2i>, boundBox: AABB, spawnParameters: SpawnParameters): Vector2d? {
		val allowAir = SpawnArea.AIR in spawnParameters.areas
		val allowSurface = SpawnArea.SURFACE in spawnParameters.areas
		val allowCeiling = SpawnArea.CEILING in spawnParameters.areas
		val allowLiquid = SpawnArea.LIQUID in spawnParameters.areas
		val allowSolid = SpawnArea.SOLID in spawnParameters.areas

		for (position in positions) {
			val region = boundBox + position

			// Original engine checks this *after* solid/liquid, which seems to be wrong.
			if (world.chunkMap.anyCellSatisfies(region.padded(Globals.spawner.spawnProhibitedCheckPadding, Globals.spawner.spawnProhibitedCheckPadding)) { _, _, cell -> cell.foreground.material.value.collisionKind == CollisionType.NULL || cell.dungeonId != NO_DUNGEON_ID })
				continue

			val isSolidSpace = world.chunkMap.anyCellSatisfies(region) { _, _, cell -> cell.foreground.material.value.collisionKind.isSolidCollision }

			if (isSolidSpace) {
				if (allowSolid)
					return position.toDoubleVector()
				else
					continue
			}

			val liquid = world.chunkMap.averageLiquidLevel(region)

			if (liquid != null && liquid.average >= Globals.spawner.minimumLiquidLevel) {
				if (allowLiquid)
					return position.toDoubleVector()
				else
					continue
			}

			if (allowAir)
				return position.toDoubleVector()
			else if (allowSurface) {
				for (sd in 0 .. Globals.spawner.spawnSurfaceCheckDistance) {
					val cell = world.chunkMap.getCell(region.centre.x.toInt(), (region.mins.y - sd).toInt())

					if (cell.foreground.material.value.collisionKind.isFloorCollision)
						return position.toDoubleVector()
				}
			} else if (allowCeiling) {
				for (sd in 0 .. Globals.spawner.spawnCeilingCheckDistance) {
					val cell = world.chunkMap.getCell(region.centre.x.toInt(), (region.mins.y + sd).toInt())

					if (cell.foreground.material.value.collisionKind.isSolidCollision)
						return position.toDoubleVector()
				}
			}
		}

		return null
	}

	// TODO: This does not support directional gravity
	private suspend fun spawnInCell(cell: Vector2i) {
		val spawnRegion = cell2region(cell)

		val spawnRegionPositions by lazy(LazyThreadSafetyMode.NONE) {
			val result = ArrayList<Vector2i>()

			for (x in spawnRegion.mins.x until spawnRegion.maxs.x) {
				for (y in spawnRegion.mins.y until spawnRegion.maxs.y) {
					result.add(Vector2i(x, y))
				}
			}

			result
		}

		val tickets = world.permanentChunkTicket(
			spawnRegion,
			ChunkState.FULL // TODO: avoid chunkloading?
		).await()

		try {
			tickets.forEach { it.chunk.await() }

			val chunkMap = world.chunkMap.snapshot()
			val spawnParameters = Starbound.EXECUTOR.supplyAsync { getSpawnParameters(cell, chunkMap) }.await()

			if (spawnParameters.areas.isEmpty())
				return

			val sampleX = world.random.nextRange(spawnRegion.mins.x, spawnRegion.maxs.x)
			val sampleY = world.random.nextRange(spawnRegion.mins.y, spawnRegion.maxs.y)
			val spawnProfile = world.template.cellInfo(sampleX, sampleY).environmentBiome?.spawnProfile ?: return

			for (spawnTypeRef in spawnProfile.spawnTypes) {
				val spawnType = spawnTypeRef.value ?: continue

				if (!spawnType.spawnParameters.isCompatible(spawnParameters))
					continue

				if (world.random.nextDouble() < spawnType.spawnChance) {
					val spawnSeed = staticRandom64(spawnType.actualSeedMix, world.template.seed)
					val targetGroupSize = world.random.nextRange(spawnType.groupSize)

					for (i in 0 until targetGroupSize) {
						val monsterType = spawnType.monsterType.map({ it.sample(world.random).orThrow { RuntimeException() } }, { it })

						if (monsterType.isEmpty) {
							LOGGER.error("Tried to spawn monster $monsterType, but it is invalid")
							continue
						}

						var variant = monsterType.value!!.create(spawnSeed, spawnType.monsterParameters)
						val monsterBoundBox = variant.commonParameters.movementSettings.standingPoly?.map({ it.aabb }, { it.stream().map { it.aabb }.reduce { t, u -> t.combine(u) }.getOrNull() })

						if (monsterBoundBox == null) {
							LOGGER.error("Tried to spawn monster $monsterType, but it has no valid AABB to check spawn conditions against")
							continue
						}

						spawnRegionPositions.shuffle(world.random)
						val position = selectSpawnPosition(spawnRegionPositions.toTypedArray(), monsterBoundBox, spawnType.spawnParameters) ?: continue
						var level = world.template.threatLevel

						if (world.sky.dayLevel >= Globals.spawner.minimumDayLevel)
							level += world.random.nextRange(spawnType.dayLevelAdjustment)
						else
							level += world.random.nextRange(spawnType.nightLevelAdjustment)

						val finalSpawnProfile = world.template.cellInfo(position.x.toInt(), position.y.toInt()).environmentBiome?.spawnProfile ?: continue

						val reUniqueParameters = mergeJson(variant.uniqueParameters.deepCopy(), finalSpawnProfile.monsterParameters)
						val reParameters = mergeJson(variant.parameters.deepCopy(), reUniqueParameters)

						variant = variant.copy(
							uniqueParameters = reUniqueParameters,
							parameters = reParameters
						)

						val entity = MonsterEntity(variant, level)
						entity.movement.position = position
						entity.isPersistent = false
						entity.joinWorld(world)
						spawnedEntities.add(entity)
					}
				}
			}
		} finally {
			tickets.forEach { it.cancel() }
			activeCells[cell] = world.simulationTime + Globals.spawner.spawnCellLifetime

			if (!hasCleanupTask) {
				hasCleanupTask = true
				world.eventLoop.schedule(::cleanupActiveCells, (Globals.spawner.spawnCellLifetime * 1000.0).toLong(), TimeUnit.MILLISECONDS)
			}
		}
	}

	private fun cleanupActiveCells() {
		val itr = activeCells.object2DoubleEntrySet().iterator()
		var min = Double.MAX_VALUE

		while (itr.hasNext()) {
			val next = itr.next()

			if (next.doubleValue <= world.simulationTime) {
				itr.remove()
			} else {
				min = min(min, next.doubleValue)
			}
		}

		if (min != Double.MAX_VALUE) {
			world.eventLoop.schedule(::cleanupActiveCells, ((min - world.simulationTime) * 1000.0).toLong(), TimeUnit.MILLISECONDS)
		} else {
			hasCleanupTask = false
		}
	}

	fun tick() {
		if (!active)
			return

		for (client in world.clients) {
			for (window in client.client.trackingTileRegions()) {
				for (splitRange in world.geometry.split(window.padded(Globals.spawner.windowActivationBorder, Globals.spawner.windowActivationBorder)).first) {
					val indexes = AABBi.of(AABB.of(splitRange) / Globals.spawner.spawnCellSize.toDouble())

					// TODO: original engine has < condition here
					//  while we have <=
					for (x in indexes.mins.x .. indexes.maxs.x) {
						for (y in indexes.mins.y .. indexes.maxs.y) {
							val cell = Vector2i(x, y)

							if (cell !in activeCells) {
								activeCells[cell] = Double.MAX_VALUE
								world.eventLoop.scope.launch { spawnInCell(cell) }
							} else {
								val existing = activeCells.getDouble(cell)

								if (existing != Double.MAX_VALUE) {
									activeCells[cell] = world.simulationTime + Globals.spawner.spawnCellLifetime
								}
							}
						}
					}
				}
			}
		}

		val itr = spawnedEntities.iterator()

		for (entity in itr) {
			if (!entity.isInWorld) {
				itr.remove()
			} else {
				val cellX = (entity.movement.xPosition / Globals.spawner.spawnCellSize).toInt()
				val cellY = (entity.movement.yPosition / Globals.spawner.spawnCellSize).toInt()

				if (Vector2i(cellX, cellY) !in activeCells) {
					entity.remove(AbstractEntity.RemovalReason.REMOVED)
					itr.remove()
				}
			}
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
